from typing import Any

from miv.core.datatype.events import *
from .pure_python import *
from .signal import *
from .spikestamps import *

DataTypes = Any  # Union[ # TODO
#    miv.core.datatype.signal.Signal,
#    miv.core.datatype.signal.Spikestamps,
# ]
